/*
 * advanced socket bouncer
 */
#include <stdio.h>
#include <unistd.h>
#include "main.h"
#include "netcode.h"
#include "socks.h"
#include "squid.h"
#include "wingate.h"

void info();
void letsdoit();
int filter(char *host);
void readconf(char *filename, struct servers **list);

int main(int argc, char **argv) {
   char execute[MAXPFAD] = "",
        configfile[MAXPFAD] = "",
   	target[MAXHOST] = "",
   	source[MAXHOST] = "";
   int  list = 0,
        optchar, 
        error = 0,      
        sourceport = 0,
        targetport = 80,
        family = -1,
        protocol = SOCK_STREAM,
        rc,
        s,
        c;
   struct hostent *hc = NULL,
                  *hv = NULL;
   struct servers *slist = NULL,
                  *prev = NULL;

   if (argc == 1) {
      info();
      return 0x99;
   } else {
      while ( (optchar = getopt(argc, argv, "6ls:e:t:c:")) != EOF ) {
         switch(optchar) {
            case 'l': list = 1; 
                      break;
            case 'c': strncpy(configfile, optarg, sizeof(configfile) - 1);
                      break;
            case 'e': strncpy(execute, optarg, sizeof(execute) - 1);
                      break;
            case 't': strncpy(target, optarg, sizeof(target) - 1);
                      break;
            case 's': strncpy(source, optarg, sizeof(target) - 1);
                      break;
#ifdef INET6
            case '6': family = AF_INET6;
                      break;
#endif 
         }
      } 
   }

   if (!list) {
      targetport = filter(target);
#ifdef INET6
      gethost(&hc, target, family);
#else
      gethost(&hc, target, AF_INET);
#endif
      if (!hc) {
         fprintf(stderr, "cant resolve target host\n");
         return 0x93;
      }
   } else {
      sourceport = filter(source); 
#ifdef INET6
      gethost(&hc, source, family);
#else
      gethost(&hc, source, AF_INET);
#endif
      if (!hc) {
         fprintf(stderr, "cant resolve source host\n");
         return 0x93;
      }
   }

   s = socket(hc->h_addrtype, protocol, IPPROTO_TCP);
   if (s == EOF) {
      fprintf(stderr, "cant create socket\n");
      return 0x91;
   }

   if (strcmp(source, "")) {
      if (!list) sourceport = filter(source);
      gethost(&hv, source, hc->h_addrtype);
      if (!hv) {
         fprintf(stderr, "cant resolve source host\n");
         return 0x90;
      } else {
         family = hv->h_addrtype;
      }
      rc = sockethandling(s, BIND, family, hv->h_addr, hv->h_length, sourceport); 
      if (rc == EOF) {
         fprintf(stderr, "bind() failed\n");
         return 0x92;
      }
   }

   if (!list) {
      rc = sockethandling(s, CONNECT, hc->h_addrtype, hc->h_addr, hc->h_length, targetport);
      if (rc == EOF) {
         fprintf(stderr, "connect() failed\n");
         return 0x94;
      }

      if (strcmp(configfile, "")) readconf(configfile, &slist);  
 
      if (slist) {
         do {
            switch(slist->type) {
               case SOCKS4: if (!socks4cmd(s, CONNECT, slist->host, slist->port, NULL)) error++;
		     	    break;
	       case SOCKS5: if (!socks5cmd(s, CONNECT, slist->host, slist->port, NULL)) error++;
                            break;
               case SQUID:  if (!squidconnect(s, slist->host, slist->port)) error++;
                            break;
               case WINGATE: if (!wingateconnect(s, slist->host, slist->port)) error++;
                             break;
            }
        
            prev = slist;
            slist = slist->next;
            free(prev);
         } while (slist && !error);
      }
      if (!error) bounce(s);
   } else {
      if (listen(s, 1) == -1) {
         fprintf(stderr," listen() = -1\n");
         return 0x95;
      }
      c = sockethandling(s, ACCEPT, hc->h_addrtype, hc->h_addr, hc->h_length, sourceport);
      dup2(c, 0);
      dup2(c, 1);
      dup2(c, 2);
      system(execute);
      close(c);
   }
 
   close(s);

   return 0;
}

void readconf(char *filename, struct servers **list) {
   FILE *fp = fopen(filename, "r");
   struct servers *p , *last;
   char type[10] = "",
   	host[MAXHOST] = "";
   int  port,
        protocol; 

   if (fp) {
      p = last = NULL;
      port = protocol = 0;
      while(fscanf(fp, "%10s %128s %d %d", type, host, &port, &protocol) > 2) {
          p = (struct servers*) malloc(sizeof(struct servers));  
          if ( *list == NULL) 
            *list = p;
          else {
            last = *list;
            while ( last->next != NULL) last = last->next;
            last->next = p;
          }
	  p->next = NULL;
          strcpy(p->host, host);
          p->port = port;
          if (protocol == 0 || protocol == 4) p->family = AF_INET;
          else if (protocol == 6) p->family = AF_INET6;
          if (!strcasecmp(type, "socks4")) p->type = SOCKS4;
          else if (!strcasecmp(type, "socks5")) p->type = SOCKS5;
          else if (!strcasecmp(type, "squid")) p->type = SQUID;
          else if (!strcasecmp(type, "wingate")) p->type = WINGATE;
      }
      fclose(fp);
   }
}


int filter(char *host) {
    char *p = (char*) strchr(host, SEPERATOR);
    int  port = 0;  

    if (p) {
       port = atoi(p + 1); 
       *p = 0;
    }

    return port;
}

void info() {
   fprintf(stderr, "%s - version %s, Wildandi <wildandi@void.at>\n", PROGRAM, VERSION); 
   fprintf(stderr, "-t host%cport ... target\n", SEPERATOR);
   fprintf(stderr, "-s host%cport ... source\n", SEPERATOR);
   fprintf(stderr, "-l ............. listen mode\n");
   fprintf(stderr, "-e program ..... execute program\n");
#ifdef INET6
   fprintf(stderr, "-6 ............. force IPv6 connections\n");
#endif 
   fprintf(stderr, "-c ............. config file that includes servers\n");
}
